/* ttt-client.c - client handling code for tic-tac-toe game server
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include "ttt-client.h"

#include "ttt-error.h"
#include "ttt-lex.h"
#include "ttt-server.h"
#include "ttt-socket.h"
#include "ttt-token.h"

struct _ttt_client {
    pthread_mutex_t mutex;
    pthread_t       thread;

    ttt_server_t    *server;
    int             socket;
    yyscan_t        scanner;

    char            **request_strings;
    int             num_request_strings;

    char            *name;
    ttt_bool_t      registered;
    int             num_wins;
};

typedef ttt_error_t (*ttt_command_func_t) (ttt_client_t *client,
					   char **args,
					   int num_args);

static ttt_error_t
_ttt_client_execute_helo (ttt_client_t *client,
			  char         **args,
			  int          num_args);

static ttt_error_t
_ttt_client_execute_who (ttt_client_t *client,
			 char         **args,
			 int          num_args);

static ttt_error_t
_ttt_client_execute_quit (ttt_client_t *client,
			  char         **args,
			  int          num_args);

static ttt_error_t
_ttt_client_execute_message (ttt_client_t *client,
			     char         **args,
			     int          num_args);



typedef struct _ttt_command_description {
    const char         *command;
    int                args_required;
    ttt_command_func_t execute;
} ttt_command_description_t;

ttt_command_description_t command_descriptions[] = {
    {"HELO",    1, _ttt_client_execute_helo},
    {"WHO",     0, _ttt_client_execute_who },
    {"QUIT",    0, _ttt_client_execute_quit},
    {"MESSAGE", 1, _ttt_client_execute_message}
};

#define ARRAY_SIZE(arr) (sizeof(arr)/sizeof(arr[0]))

static ttt_error_t
_ttt_client_execute_helo (ttt_client_t *client,
			  char         **args,
			  int          num_args)
{
    ttt_error_t error;
    char *response;

    assert (num_args == 1);

    ttt_client_set_name (client, args[0]);

    error = ttt_server_register_client (client->server, client);
    if (error)
	return error;

    client->registered = TRUE;

    xasprintf (&response, "HELO %s %s %s\r\n",
	       client->name,
	       ttt_server_get_host (client->server),
	       ttt_server_get_port (client->server));

    ttt_client_send (client, response);

    xasprintf (&response, "NOTICE USER %s\r\n",client->name);
    ttt_server_broadcast(client->server,response);

    free (response);

    return TTT_ERROR_NONE;
}

static ttt_error_t
_ttt_client_execute_who (ttt_client_t *client,
			 char         **args,
			 int          num_args)
{
    char *response;

    assert (num_args == 0);

    response = ttt_server_who (client->server);

    ttt_client_send (client, response);

    free (response);

    return TTT_ERROR_NONE;
}

static ttt_error_t
_ttt_client_execute_quit (ttt_client_t *client,
			  char	       **args,
			  int	       num_args)
{
    char *notice;

    assert (num_args == 0);
    xasprintf (&notice,"NOTICE QUIT %s\r\n",client->name);
    ttt_server_broadcast(client->server,notice);
    free (notice);
    return TTT_ERROR_QUIT_REQUESTED;
}

static ttt_error_t
_ttt_client_execute_message (ttt_client_t  *client,
			     char	   **args,
			     int	   num_args)
{
    char *response;
    int i;

    assert (num_args > 0);

    xasprintf(&response, "MESSAGE\r\n");
    ttt_client_send (client, response);
    
    xasprintf(&response, "NOTICE MESSAGE %s", client->name);
    for (i=0; i<num_args; i++){
	xasprintf(&response, "%s %s",response,args[i]);

    }
    xasprintf(&response, "%s\r\n",response);
    ttt_server_broadcast (client->server,response);

    free (response);
    return TTT_ERROR_NONE;


}

static void
_free_request (ttt_client_t *client);

static void
_ttt_client_init (ttt_client_t	*client,
		  ttt_server_t	*server,
		  int		 socket)
{
    FILE *file;

    pthread_mutex_init (&client->mutex, NULL);

    client->server = server;
    client->socket = socket;

    file = xfdopen (socket, "r");
    yylex_init(&client->scanner);
    yyset_in (file, client->scanner);

    client->request_strings = NULL;
    client->num_request_strings = 0;

    client->name = NULL;
    client->registered = FALSE;
    client->num_wins = 0;
}

static void
_ttt_client_fini (ttt_client_t *client)
{
    pthread_mutex_lock (&client->mutex);

    if (client->registered)
	ttt_server_unregister_client (client->server, client);

    free (client->name);
    client->name = NULL;

    yylex_destroy (client->scanner);
    shutdown (client->socket, SHUT_RDWR);

    _free_request (client);

    pthread_mutex_unlock (&client->mutex);

    pthread_mutex_destroy (&client->mutex);
}

/* XXX: The memory management for the request strings is pretty cheesy. */
static void
_append_to_request (ttt_client_t	*client,
		    const char		*string)
{
    client->num_request_strings++;
    client->request_strings =
	xrealloc (client->request_strings,
		  client->num_request_strings * sizeof (char *));

    client->request_strings[client->num_request_strings - 1] = xstrdup (string);
}

static void
_free_request (ttt_client_t *client)
{
    int i;

    for (i = 0; i < client->num_request_strings; i++)
	free (client->request_strings[i]);

    free (client->request_strings);

    client->request_strings = NULL;
    client->num_request_strings = 0;
}

static ttt_status_t
_read_request (ttt_client_t *client)
{
    ttt_token_t token;

    _free_request (client);

    while (1) {
	token = yylex (client->scanner);
	/* Yes, EOF in two different enums is pretty ugly. */
	if (token == TTT_TOKEN_EOF)
	    return TTT_STATUS_EOF;

	if (token == TTT_TOKEN_NEWLINE) {
	    if (client->num_request_strings)
		return TTT_STATUS_SUCCESS;
	    else
		continue;
	}

	assert (token == TTT_TOKEN_STRING);

	_append_to_request (client, yyget_text (client->scanner));
    }
}

static ttt_error_t
_execute_request (ttt_client_t *client)
{
    int i;

    char *command = client->request_strings[0];
    int num_args = client->num_request_strings-1;
    ttt_command_description_t *desc;

    for (i=0; i < strlen (command); i++)
	command[i] = toupper (command[i]);

    for (i=0; i < ARRAY_SIZE(command_descriptions); i++) {
	desc = &command_descriptions[i];
	if (strcmp(command, desc->command) == 0) {
	    if (num_args < desc->args_required)
		return TTT_ERROR_SYNTAX;
	    return (desc->execute) (client, &client->request_strings[1], num_args);
	    }
	}

    return TTT_ERROR_COMMAND;
}

static void *
_handle_requests_thread (void *closure)
{
    ttt_status_t status;
    ttt_error_t error;
    ttt_client_t *client = closure;

    while (1) {

	status = _read_request (client);
	if (status == TTT_STATUS_EOF)
	    break;
	if (status)
	    ASSERT_NOT_REACHED;

	error = _execute_request (client);
	if (error == TTT_ERROR_QUIT_REQUESTED)
	    break;
	if (error)
	    ttt_client_send (client, ttt_error_string (error));
    }

    _ttt_client_fini (client);
    free (client);

    return (void *) 0;
}

/* Exported: See ttt-client.h for documentation. */
void
ttt_client_new (void *closure, int client_socket)
{
    ttt_server_t *server = closure;
    ttt_client_t *client;
    int err;

    client = xmalloc (sizeof (ttt_client_t));
    
    _ttt_client_init (client, server, client_socket);

    err = pthread_create (&client->thread, NULL,
			  _handle_requests_thread, client);
    if (err != 0) {
	fprintf (stderr, "Error: pthread_create failed: %s. Aborting.\r\n",
		 strerror (err));
	exit (1);
    }
}

/* Exported: See ttt-client.h for documentation. */
void
ttt_client_send (ttt_client_t *client, const char *message)
{
    pthread_mutex_lock (&client->mutex);

    ttt_socket_write (client->socket, message, strlen (message));

    pthread_mutex_unlock (&client->mutex);
}

/* Exported: See ttt-client.h for documentation. */
const char*
ttt_client_get_name (ttt_client_t *client)
{
    return client->name;
}

/* Exported: See ttt-client.h for documentation. */
void
ttt_client_set_name (ttt_client_t *client, const char *name)
{
    free (client->name);
    client->name = xstrdup (name);
}

/* Exported: See ttt-client.h for documentation. */
int
ttt_client_get_num_wins (ttt_client_t *client)
{
    return client->num_wins;
}
