/* ttt-server.c - tic-tac-toe game server
 *
 * Copyright © 2005 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include "ttt-server.h"

#include "ttt-args.h"
#include "ttt-client.h"
#include "ttt-error.h"
#include "ttt-socket.h"

struct _ttt_server {
    pthread_mutex_t mutex;

    const char *host;
    const char *port;

    ttt_client_t **clients;
    int clients_size;
    int num_clients;
};

static void
ttt_server_init (ttt_server_t *server, const char *host, const char *port)
{
    pthread_mutex_init (&server->mutex, NULL);

    server->host = host;
    server->port = port;

    server->clients = NULL;
    server->clients_size = 0;
    server->num_clients = 0;
}

/* Exported: See ttt-server.h for documentation. */
ttt_error_t
ttt_server_register_client (ttt_server_t *server, ttt_client_t *client)
{
    int i;
    ttt_error_t error = TTT_ERROR_NONE;
    const char *name;

    pthread_mutex_lock (&server->mutex);

    name = ttt_client_get_name (client);

    assert (name != NULL);

    if (name[0] == '\0')
	return TTT_ERROR_INVALID_NAME;

    for (i = 0; i < server->num_clients; i++) {
	if (strcmp (ttt_client_get_name (server->clients[i]), name) == 0) {
	    error = TTT_ERROR_INVALID_NAME;
	    goto CLEANUP_LOCK;
	}
    }

    printf ("Client %s has joined.\r\n", name);

    server->num_clients++;

    if (server->num_clients > server->clients_size) {
	if (server->clients_size == 0)
	    server->clients_size = 1;
	else
	    server->clients_size *= 2;

	server->clients = xrealloc (server->clients,
				    server->clients_size * sizeof (ttt_client_t *));
    }

    server->clients [server->num_clients - 1] = client;

 CLEANUP_LOCK:
    pthread_mutex_unlock (&server->mutex);

    return error;
}

/* Exported: See ttt-server.h for documentation. */
void
ttt_server_unregister_client (ttt_server_t *server, ttt_client_t *client)
{
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++)
	if (server->clients[i] == client)
	    break;

    assert (i < server->num_clients);

    printf ("Client %s has left.\r\n", ttt_client_get_name (client));

    memmove (&server->clients[i], &server->clients[i+1],
	     (server->num_clients - i - 1) * sizeof (ttt_client_t *));

    server->num_clients--;

    pthread_mutex_unlock (&server->mutex);
}

/* Exported: See ttt-server.h for documentation. */
void
ttt_server_broadcast (ttt_server_t *server, const char *message)
{
    int i;

    pthread_mutex_lock (&server->mutex);

    for (i = 0; i < server->num_clients; i++)
	ttt_client_send (server->clients[i], message);

    pthread_mutex_unlock (&server->mutex);
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_who (ttt_server_t *server)
{
    int i;
    char *response;

    pthread_mutex_lock (&server->mutex);

    xasprintf (&response, "WHO");

    for (i = 0; i < server->num_clients; i++)
	xasprintf (&response, "%s %s %d",
		   response,
		   ttt_client_get_name (server->clients[i]),
		   ttt_client_get_num_wins (server->clients[i]));

    xasprintf (&response, "%s\r\n", response);

    pthread_mutex_unlock (&server->mutex);

    return response;
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_get_host (ttt_server_t *server)
{
    return server->host;
}

/* Exported: See ttt-server.h for documentation. */
const char*
ttt_server_get_port (ttt_server_t *server)
{
    return server->port;
}

static const char *WELCOME_MESSAGE = 
"Welcome to ttt-server. So far, this program is still a demonstration\r\n"
"TCP/IP server, acting something like a rather braindead chat server.\r\n"
"The server is currently listening on:\r\n"
"\r\n	%s:%s\r\n"
"\r\nTo test this, simply connect one or more clients to that host and port.\r\n"
"For example:\r\n"
"\r\n	telnet %s %s\r\n"
"\r\nOnce you have connected a client, the server will send each line of text\r\n"
"it receives to all connected clients. The server reports client joins and\r\n"
"departures on stdout.\r\n"
"\r\nNote that to terminate the telnet client you type Control-], then\r\n"
"<Enter>, then \"close\" (and <Enter>) at the \"telnet> \" prompt.\r\n"
"\r\nHave fun!\r\n"
"-Carl\r\n"
"\r\nPS. At this point we're ready to leave the demonstration phase and to\r\n"
"begin implementing TTTP (tic-tac-toe protocol) as well as fixing the\r\n"
"protocol specifcation. We don't need a custom client to move forward on\r\n"
"the server (that is one of the ideas behind using a telnet-compatible\r\n"
"protocol), but a custom client would still be a fine project for a\r\n"
"motivated beginning programmer.\r\n\r\n";

static void
_ttt_server_accept (void *closure, int client_socket)
{
    ttt_server_t *server = closure;

    ttt_client_new (server, client_socket);
}

int 
main (int argc, char **argv)
{
    ttt_args_t args;
    ttt_server_t server;
    int socket;

    ttt_args_parse (&args, argc, argv);

    if (args.log_file)
	xfreopen (args.log_file, "a", stderr);

    socket = ttt_socket_create_server (args.host, args.port);

    printf (WELCOME_MESSAGE, args.host, args.port, args.host, args.port);

    ttt_server_init (&server, args.host, args.port);

    while (1)
	ttt_socket_accept (socket, _ttt_server_accept, &server);

    /* We only reach here if something bad happened. */
    return 1;
}
